//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

import SafariServices
import Alamofire

import KakaoSDKCommon
import KakaoSDKAuth
import KakaoSDKUser
import KakaoSDKTalk
import KakaoSDKShare
import KakaoSDKNavi
import KakaoSDKTemplate

import KakaoPartnerSDKCommon
import KakaoPartnerSDKUser
import KakaoPartnerSDKTalk
import KakaoPartnerSDKAuth
import KakaoPartnerSDKShare

import KakaoSDKFriend
import KakaoPartnerSDKFriend
import KakaoPartnerSDKFriendDelegate

extension PartnerSDKSampleViewController {
    private func updateResultIndicator(cell:UITableViewCell?, row:Row) {
        
        switch(row.apiResult) {
        case .Ready: cell?.backgroundColor = UIColor.white
        case .Success: cell?.backgroundColor = UIColor.green
        case .Fail: cell?.backgroundColor = UIColor.red
        case .NotSuccess: cell?.backgroundColor = UIColor.red
        }
    }
    
    private func updateResultIndicator() {
        if let resultIndicator = self.resultIndicator {
            let cell = self.tableView.cellForRow(at: resultIndicator.0)
            self.updateResultIndicator(cell: cell, row: resultIndicator.1)
        }
    }
    
    private func success(_ object: Any? = nil, updateIndicator: Bool = true) {
        if let object = object {
            print("\(String(describing:object))")
        }
        
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = .Success
                self.updateResultIndicator()
            }
        }
    }
    
    private func errorHandler(error:Error?, updateIndicator: Bool = true, withAlert: Bool = false) {
        guard error != nil else { return }
        if (updateIndicator) {
            DispatchQueue.main.async {
                self.resultIndicator?.1.apiResult = .Fail
                self.updateResultIndicator()
            }
        }
        
        if let sdkError = error as? SdkError {
            print("\n\n[sdk error handling start]------------------------------------------------------------------------------")
            
            switch(sdkError) {
            case .ApiFailed:
                let apiError = sdkError.getApiError()
                switch(apiError.reason) {
                //if need, use reason case
                default:
                    print("api error: \(String(describing: apiError))")
                }
            case .AuthFailed:
                let authError = sdkError.getAuthError()
                switch(authError.reason) {
                //if need, use reason case
                default:
                    print("auth error: \(String(describing: authError))")
                }
            case .ClientFailed:
                let clientError = sdkError.getClientError()
                switch(clientError.reason) {
                //if need, use reason case
                default:
                    print("client error: \(String(describing: clientError))")
                }
            case .AppsFailed:
                let appsError = sdkError.getAppsError()
                switch(appsError.reason) {
                //if need, use reason case
                default:
                    print("apps error: \(String(describing: appsError))")
                }
            @unknown default:
                break
            }
            
            if withAlert == true {
                self.showAlert(title:"SDK Error", message: String(describing: error))
            }
            print("[sdk error handling end]--------------------------------------------------------------------------------\n")
        }
        else {
            print("\n\n[not sdk error start]------------------------------------------------------------------------------")
            if let nsError = error as NSError? {
                print("normal error: \(String(describing: nsError))")
                
                if withAlert == true {
                    self.showAlert(title:"Normal(Not SDK) Error", message: String(describing: nsError))
                }
            }
            
            print("[not sdk error end]------------------------------------------------------------------------------\n")
        }
    }
}


class PartnerSDKSampleViewController: SDKSampleBaseViewController, UITableViewDelegate, UITableViewDataSource {
    
    var safariViewController : SFSafariViewController?
    let paramDialogRequestColor : UIColor = UIColor.init(red: 0, green: 1, blue: 1, alpha: 0.5)
    
    var recursiveAppPartnerFriendsCompletion : ((Friends<PartnerFriend>?, Error?) -> Void)!
    
    override func loadView() {
        super.loadView()
        
        setupViews()
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        self.tableView.delegate = self
        self.tableView.dataSource = self
        
        setMenu()
    }
    
    override func setupViews() {
        super.setupViews()
        self.tableView.register(UITableViewCell.self, forCellReuseIdentifier: "Cell")
    }
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        let item = self.sections![indexPath.section].items[indexPath.row]
        if let action = item.action {
            action()
            
            self.resultIndicator = (indexPath, item)
            
            DispatchQueue.main.asyncAfter(deadline: .now() + 0.1 * 1) {
               tableView.deselectRow(at: indexPath, animated: false)
            }
        }
    }
    
    func numberOfSections(in tableView: UITableView) -> Int {
        if let sections = self.sections, sections.count != 0 {
            print("sections count: ", sections.count)
            return sections.count
        }
        else {
            return 0
        }
    }
    
    func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        if let sections = self.sections, sections.count != 0 {
            return sections[section].header
        }
        else {
            return nil
        }
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if let sections = self.sections, sections.count != 0 {
            return sections[section].items.count
        }
        else {
            return 0
        }
    }

    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "Cell", for: indexPath)
        
        let item = self.sections![indexPath.section].items[indexPath.row]
        
        self.updateResultIndicator(cell:cell, row:item)
        cell.textLabel?.adjustsFontSizeToFitWidth = true
        cell.textLabel?.text = "\(item.title)"
        cell.textLabel?.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
        cell.textLabel?.textColor = UIColor.black
        
        if let backgroundColor = item.backgroundColor {
            cell.backgroundColor = backgroundColor
        }
        
        return cell
    }
    
    func setTitle(title:String = "SDK Sample", args: String? = nil) {
        self.navigationItem.title = (args != nil) ? "\(title)(\(args!))" : "\(title)"
    }
    
    func extractPickerElements(friends:Friends<PartnerFriend>, receiverIdType:TalkApi.ReceiverIdType = .Uuid) -> [(URL?, String?, Any)]? {
        return friends.elements?.map({ (friend) -> (URL?, String?, Any) in
            
            switch (receiverIdType) {
            case .UserId:
                return (friend.profileThumbnailImage, friend.profileNickname, "\(friend.id ?? 0)")
            case .Uuid:
                return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid)
            default:
                return (friend.profileThumbnailImage, friend.profileNickname, friend.uuid)
            }
        })
    }
    
    func extractPickerElements(chats:Chats) -> [(URL?, String?, Any)]? {
        return chats.elements?.map({ (chat) -> (URL?, String?, Any) in
            
            var chatImageUrl : URL? = nil
            if let imageUrl = chat.imageUrl {
                chatImageUrl = imageUrl
            }
            
            if let displayMemberImageUrl = chat.displayMemberImages?[0] {
                chatImageUrl = displayMemberImageUrl
            }
            
            return (chatImageUrl, chat.title, chat.id)
        })
    }
    
    func getMostTopViewController(base: UIViewController? = UIApplication.sdkKeyWindow()?.rootViewController) -> UIViewController? {

        if let naviController = base as? UINavigationController {
            return getMostTopViewController(base: naviController.visibleViewController)

        } else if let tabbarController = base as? UITabBarController, let selected = tabbarController.selectedViewController {
            return getMostTopViewController(base: selected)

        } else if let presented = base?.presentedViewController {
            return getMostTopViewController(base: presented)
        }
        return base
    }
    
    
    
    func setMenu() {
        self.sections = [
            // MARK: - section FRIEND start =============================================================================================
            SectionOfRow(header: "FRIEND", items: [
                Row(title: "PickerApi launcher - Tab", action: {
                    let vc = PartnerSDKFriendPickerSettingViewController()
                    vc.viewModel = PartnerSDKTapPickerSettingViewModel()
                    vc.modalPresentationStyle = .fullScreen
                    self.present(vc, animated: true)
                }, backgroundColor: .orange),
                Row(title: "select(params:) - .chat", action: {
                    let friendRequestParams = PickerTabRequestParams.FriendRequestParams(friendFilter: nil,                //default PickerFriendFilter.none
                                                                                         countryCodeFilters: nil,                      //default nil
                                                                                         usingOsFilter: .all,                    //default .all
                                                                                         enableIndex: true,                      //default true
                                                                                         showMyProfile: true,                    //default true
                                                                                         showFavorite: true,                     //default true
                                                                                         showPickedFriend: nil)                  //default true
                                                                                         
                    let chatRequestParams = PickerTabRequestParams.ChatRequestParams(selectionType:.chat,              //required  .chat or .chatMember
                                                                                     chatFilters: [.regular])    //default [.regular]
                    
                    let tabRequestParams = PickerTabRequestParams(title: nil,                                           //default "친구 선택"
                                                                  viewAppearance: .auto,                                //default .auto
                                                                  orientation: .auto,                                   //default .auto
                                                                  enableSearch: true,                                   //default true
                                                                  disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],  //defalut nil
                                                                  displayAllProfile: nil,                               //default false
                                                                  maxPickableCount: nil,                  //default 30
                                                                  minPickableCount: nil,
                                                                  friendParams: friendRequestParams,
                                                                  chatParams: chatRequestParams)
                    
                    PickerApi.shared.select(params: tabRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectPopup(params:) - .chat", action: {
                    let friendRequestParams = PickerTabRequestParams.FriendRequestParams(friendFilter: nil,                //default PickerFriendFilter.none
                                                                                         countryCodeFilters: nil,                      //default nil
                                                                                         usingOsFilter: .all,                    //default .all
                                                                                         enableIndex: true,                      //default true
                                                                                         showMyProfile: true,                    //default true
                                                                                         showFavorite: true,                     //default true
                                                                                         showPickedFriend: nil)                  //default true
                                                                                         
                    let chatRequestParams = PickerTabRequestParams.ChatRequestParams(selectionType:.chat,              //required  .chat or .chatMember
                                                                                     chatFilters: [.regular])          //default [.regular]
                    
                    let tabRequestParams = PickerTabRequestParams(title: nil,                                                   //default "친구 선택"
                                                                  viewAppearance: .auto,                                        //default .auto
                                                                  orientation: .auto,                                           //default .auto
                                                                  enableSearch: true,                                           //default true
                                                                  disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],    //defalut nil
                                                                  displayAllProfile: nil,                                       //default false
                                                                  maxPickableCount: nil,                                        //default 30
                                                                  minPickableCount: nil,
                                                                  friendParams: friendRequestParams,
                                                                  chatParams: chatRequestParams)
                    
                    PickerApi.shared.selectPopup(params: tabRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "select(params:) - .chatMember", action: {
                    let friendRequestParams = PickerTabRequestParams.FriendRequestParams(friendFilter: nil,                //default PickerFriendFilter.none
                                                                                         countryCodeFilters: nil,                      //default nil
                                                                                         usingOsFilter: .all,                    //default .all
                                                                                         enableIndex: true,                      //default true
                                                                                         showMyProfile: true,                    //default true
                                                                                         showFavorite: true,                     //default true
                                                                                         showPickedFriend: nil)                  //default true
                                                                                         
                    let chatRequestParams = PickerTabRequestParams.ChatRequestParams(selectionType:.chatMember,         //required  .chat or .chatMember
                                                                                     chatFilters: [.regular])     //default [.regular]
                    
                    let tabRequestParams = PickerTabRequestParams(title: nil,                                           //default "친구 선택"
                                                                  viewAppearance: .auto,                                //default .auto
                                                                  orientation: .auto,                                   //default .auto
                                                                  enableSearch: true,                                   //default true
                                                                  disableSelectOptions: [DisableSelectOption(reason:.msgBlocked),
//                                                                                   DisableSelectOption(reason:.registered),
//                                                                                   DisableSelectOption(reason:.unRegistered),
//                                                                                   DisableSelectOption(reason:.custom, message: "uuid제제 테스트", uuids:["aaaaaaaaaaaaaaaa"])
                                                                  ],
                                                                  //defalut nil
                                                                  displayAllProfile: nil,                               //default false
                                                                  maxPickableCount: nil,                                //default 30
                                                                  minPickableCount: nil,
                                                                  friendParams: friendRequestParams,
                                                                  chatParams: chatRequestParams)
                    
                    PickerApi.shared.select(params: tabRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                
                Row(title: "selectPopup(params:) - .chatMember", action: {
                    let friendRequestParams = PickerTabRequestParams.FriendRequestParams(friendFilter: nil,                //default PickerFriendFilter.none
                                                                                         countryCodeFilters: nil,                      //default nil
                                                                                         usingOsFilter: .all,                    //default .all
                                                                                         enableIndex: true,                      //default true
                                                                                         showMyProfile: true,                    //default true
                                                                                         showFavorite: true,                     //default true
                                                                                         showPickedFriend: nil)                  //default true
                                                                                         
                    let chatRequestParams = PickerTabRequestParams.ChatRequestParams(selectionType:.chatMember,         //required  .chat or .chatMember
                                                                                     chatFilters: [.regular])     //default [.regular]
                    
                    let tabRequestParams = PickerTabRequestParams(title: nil,                                           //default "친구 선택"
                                                                  viewAppearance: .auto,                                //default .auto
                                                                  orientation: .auto,                                   //default .auto
                                                                  enableSearch: true,                                   //default true
                                                                  disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],  //defalut nil
                                                                  displayAllProfile: nil,                               //default false
                                                                  maxPickableCount: nil,                                //default 30
                                                                  minPickableCount: nil,
                                                                  friendParams: friendRequestParams,
                                                                  chatParams: chatRequestParams)
                    
                    PickerApi.shared.selectPopup(params: tabRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                
                Row(title: "PickerApi launcher - Chat", action: {
                    let vc = PartnerSDKFriendPickerSettingViewController()
                    vc.viewModel = PartnerSDKChatPickerSettingViewModel()
                    vc.modalPresentationStyle = .fullScreen
                    self.present(vc, animated: true)
                }, backgroundColor: .orange),
                Row(title: "selectChat(params:) - .chat", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chat,                    //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          chatFilters: [.regular, .open],   //default [.regular]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true)                     //default true
                                                                          
                    
                    PickerApi.shared.selectChat(params: pickerChatRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectChatPopup(params:) - .chat", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chat,                    //required  .chat or .chatMember
                                                                          title: nil,                             //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          chatFilters: [.regular, .open],   //default [.regular]
                                                                          viewAppearance: .auto,                  //default .auto
                                                                          orientation: .auto,                     //default .auto
                                                                          enableSearch: true)                     //default true
                    
                    PickerApi.shared.selectChatPopup(params: pickerChatRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectChatMemebers(params:) - .chatMember", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chatMember,                            //required  .chat or .chatMember
                                                                          title: nil,                                           //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          chatFilters: [.regular],                        //default [.regular]
                                                                          viewAppearance: .auto,                                //default .auto
                                                                          orientation: .auto,                                   //default .auto
                                                                          enableSearch: true,                                   //default true
                                                                          disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],  //defalut nil
                                                                          displayAllProfile: nil,                               //default false
                                                                          maxPickableCount: nil,                                //default 30
                                                                          minPickableCount: nil)                                //default 1
                    
                    PickerApi.shared.selectChat(params: pickerChatRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectChatMemebersPopup(params:) - .chatMember", action: {
                    let pickerChatRequestParams = PickerChatRequestParams(selectionType:.chatMember,                            //required  .chat or .chatMember
                                                                          title: nil,                                           //default .chat:"채팅방 선택",  .chatMember:"친구 선택"
                                                                          chatFilters: [.regular],                        //default [.regular]
                                                                          viewAppearance: .auto,                                //default .auto
                                                                          orientation: .auto,                                   //default .auto
                                                                          enableSearch: true,                                   //default true
                                                                          disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],  //defalut nil
                                                                          displayAllProfile: nil,                               //default false
                                                                          maxPickableCount: nil,                                //default 30
                                                                          minPickableCount: nil)                                //default 1
                    
                    PickerApi.shared.selectChatPopup(params: pickerChatRequestParams) { [weak self] (selectedUsers, selectedChat, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "PickerApi launcher - Friend", action: {
                    let vc = PartnerSDKFriendPickerSettingViewController()
                    vc.modalPresentationStyle = .fullScreen
                    self.present(vc, animated: true)
                }, backgroundColor: .orange),
                Row(title: "selectFriends(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams(title: "샘플앱 친구피커",                     //default "카카오톡 친구 선택"
                                                                                friendFilter: nil,                            //default PickerFriendFilter.none
                                                                                countryCodeFilters: nil,                                  //default nil
                                                                                usingOsFilter: .all,                                //default .all
                                                                                viewAppearance: .auto,                              //default .auto
                                                                                orientation: .auto,                                 //default .auto
                                                                                enableSearch: true,                                 //default true
                                                                                enableIndex: true,                                  //default true
                                                                                showMyProfile: true,                                //default true
                                                                                showFavorite: true,                                 //default true
                                                                                disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],//defalut nil
                                                                                displayAllProfile: nil,                             //default false
                                                                                showPickedFriend: nil,                              //default true
                                                                                maxPickableCount: nil,                              //default 30
                                                                                minPickableCount: nil)                              //default 1
                    
                    PickerApi.shared.selectFriends(params: pickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriendsPopup(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams()
                    
                    PickerApi.shared.selectFriendsPopup(params: pickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriend(params:)", action: {
                    
                    let pickerFriendRequestParams = PickerFriendRequestParams(title: "샘플앱 친구피커",                    //default "카카오톡 친구 선택"
                                                                    friendFilter: nil,                            //default PickerFriendFilter.none
                                                                    countryCodeFilters: nil,                                  //default nil
                                                                    usingOsFilter: .all,                                //default .all
                                                                    viewAppearance: .auto,                              //default .auto
                                                                    orientation: .auto,                                 //default .auto
                                                                    enableSearch: true,                                 //default true
                                                                    enableIndex: true,                                  //default true
                                                                    showMyProfile: true,                                //default true
                                                                    showFavorite: true,                                 //default true
                                                                    disableSelectOptions: [DisableSelectOption(reason:.msgBlocked)],//defalut nil
                                                                    displayAllProfile: nil,                             //default false
                                                                    showPickedFriend: nil,                              //ignore
                                                                    maxPickableCount: nil,                              //ignore
                                                                    minPickableCount: nil)                              //ignore
                    
                    PickerApi.shared.selectFriend(params: pickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "selectFriendPopup(params:)", action: {
                    let pickerFriendRequestParams = PickerFriendRequestParams()
                    
                    PickerApi.shared.selectFriendPopup(params: pickerFriendRequestParams) { [weak self] selectedUsers, error in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
            ]),
            // section FRIEND start =============================================================================================
            
            // MARK: - section Picker Emebeded View
            SectionOfRow(header: "Embeded View", items: [
                Row(title: "Single Friend PickerView", action: { [weak self] in
                    let singleFriendEmbededViewController = SingleFriendEmbededViewController()
                    self?.present(singleFriendEmbededViewController, animated: true)
                }),
                Row(title: "Multi Friend PickerView", action: { [weak self] in
                    let multiFriendEmbededViewController = MultiFriendEmbededViewController()
                    self?.present(multiFriendEmbededViewController, animated: true)
                }),
                Row(title: "Chat Member PickerView", action: { [weak self] in
                    let chatMemberEmbededViewController = ChatMemberEmbededViewController()
                    self?.present(chatMemberEmbededViewController, animated: true)
                }),
                Row(title: "ChatMembers with masking nickname", action: { [weak self] in
                    TalkApi.shared.chatList(filters: [.Regular]) { [weak self] chats, error in
                        if let error {
                            self?.errorHandler(error: error)
                            return
                        }
                        
                        if let firstChat = chats?.elements?.first {
                            let targetInfo = TargetInfo(appKey: (try? KakaoSDK.shared.appKey()) ?? "",
                                                        bundleId: Bundle.main.bundleIdentifier ?? "",
                                                        accessToken: TokenManager.manager.getToken()?.accessToken ?? "")
                            PickerApi.shared.chatMembers(chatId: firstChat.id, targetInfo: targetInfo) { members, error in
                                if let error {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                
                                if let members {
                                    self?.success(members)
                                }
                            }
                        }
                    }
                })
            ]),
            
            // MARK: - section auth start =============================================================================================
            SectionOfRow(header: "AUTH", items: [

            ]),
            // section auth end =============================================================================================
            
            // MARK: - section user start =============================================================================================
            SectionOfRow(header: "USER", items: [
                Row(title: "loginWithKakaoAccounts(accountsSkipIntro:true)", action: {
                    UserApi.shared.loginWithKakaoAccount(accountsSkipIntro: true, completion: { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("oauthToken: \(String(describing: oauthToken))")
                            self?.success()
                        }
                    })
                }),
                Row(title: "loginWithKakaoAccounts(accountsSkipIntro:false)", action: {
                    UserApi.shared.loginWithKakaoAccount(accountsSkipIntro: false, completion: { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("oauthToken: \(String(describing: oauthToken))")
                            self?.success()
                        }
                    })
                }),
                Row(title: "loginWithKakaoAccounts(accountsTalkLoginVisible:false)", action: {
                    UserApi.shared.loginWithKakaoAccount(accountsTalkLoginVisible: false, completion: { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("oauthToken: \(String(describing: oauthToken))")
                            self?.success()
                        }
                    })
                }),
                Row(title: "loginWithKakaoAccounts(accountsTalkLoginVisible:true)", action: {
                    UserApi.shared.loginWithKakaoAccount(accountsTalkLoginVisible: true, completion: { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("oauthToken: \(String(describing: oauthToken))")
                            self?.success()
                        }
                    })
                }),
                Row(title: "loginWithKakaoAccounts(prompts:)", action: {
                    UserApi.shared.loginWithKakaoAccount(prompts: [.UnifyDaum], completion: { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("oauthToken: \(String(describing: oauthToken))")
                            self?.success()
                        }
                    })
                }),
                Row(title: "meForPartner()", action: {
                    UserApi.shared.meForPartner { [weak self] (user, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                    
//                    UserApi.shared.meForPartner(propertyKeys:["id"]) { [weak self] (user, error) in
//                        if let error = error {
//                            self?.errorHandler(error: error)
//                        }
//                        else {
//                            self?.success()
//                        }
//                    }
                }),
                Row(title: "signupForPartner()", action: {
                    UserApi.shared.signupForPartner { [weak self] (userId, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                Row(title: "+ageAuthInfo(...)", action: {
                    self.showParamDialog(apiKey: ApiKey.ageAuth) { paramItems in
                        UserApi.shared.ageAuthInfo(ageLimit: paramItems.requestParam(key:"ageLimit"),
                                                   ageCriteria: paramItems.requestParam(key:"ageCriteria"),
                                                   propertyKeys: paramItems.requestParam(key:"propertyKeys"),
                                                   completion: { [weak self] (ageAuthInfo, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                print("ageAuthInfo: \(String(describing: ageAuthInfo))")

                                self?.success()
                            }
                        })
                    }
                }, backgroundColor: paramDialogRequestColor),
                Row(title: "upgradeScopes(scopes:) - friends", action: {
                    UserApi.shared.upgradeScopes(scopes:["friends"], guardianToken: nil, completion: { [weak self] (scopeInfo, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            print("scopeInfo: \(String(describing: scopeInfo))")
                            
                            self?.success()
                        }
                    })
                }),
                Row(title: "verifyAuthAge(authLevel:ageLimit:skipTerms:authFrom:adultsOnly:underAge:)", action: {
                    UserApi.shared.verifyAge(authLevel: .Level1,
                                                    ageLimit: 30,
                                                    skipTerms: false,
                                                    adultsOnly: false,
                                                    underAge: false) { [weak self] (error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                        }
                    }
                }),
                
                Row(title: "createAddress()", action: {
                    UserApi.shared.createShippingAddress { [weak self] (addressId, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        } else {
                            self?.success(addressId)
                        }
                    }
                }),
                
                Row(title: "updateAddress(addressId:)", action: {
                    self.showParamDialog(apiKey: ApiKey.addressId) { paramItems in
                        guard let addressId: Int64 = paramItems.requestParam(key: "addressId") else { return }
                        
                        UserApi.shared.updateShippingAddress(addressId: addressId) { [weak self] (addressId, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            } else {
                                self?.success(addressId)
                            }
                        }
                    }
                })
            ]),
            // section user end =============================================================================================
            
            // MARK: - section talk start =============================================================================================
            SectionOfRow(header: "TALK", items: [
                Row(title: "chatList(offset:limit:)", action: {
                    TalkApi.shared.chatList(offset: 1, limit: 3, completion: { [weak self](chatList, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            _ = chatList
                            //do something
                            
                            self?.success(chatList)
                        }
                    })
                }),
                Row(title: "chatList(filters:offset:limit:)", action: {
                    TalkApi.shared.chatList(filters:[.Direct, .Multi, .Open, .Regular], completion: { [weak self](chatList, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            _ = chatList
                            //do something
                            
                            self?.success(chatList)
                        }
                    })
                }),
                Row(title: "chatMembers(chatId:friendsOnly:)", action: {
                    TalkApi.shared.chatList { [weak self](chatList, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            if let chats = chatList?.elements?.first {
                                TalkApi.shared.chatMembers(chatId: chats.id, friendsOnly: false) { [weak self] (chatMembers, error) in
                                    if let error = error {
                                        self?.errorHandler(error: error)
                                    }
                                    else {
                                        self?.success()
                                        
                                        _ = chatMembers
                                        //do something
                                    }
                                }
                            }
                            else {
                                self?.errorHandler(error:SampleError("chats empty!"))
                            }
                        }
                    }
                }),
                Row(title: "friendsForPartner()", action: {
                    TalkApi.shared.friendsForPartner { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            _ = friends
                            //do something
                        }
                    }
                }),
                Row(title: "friendsForPartner(countryCodes:)", action: {
                    TalkApi.shared.friendsForPartner(countryCodes:["kr","us"], completion:{ [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            _ = friends
                            //do something
                        }
                    })
                }),
                Row(title: "friendsForPartner(context:) - recursive", action: {
                    var nextPartnerFriendContext : PartnerFriendsContext? = PartnerFriendsContext(offset: 0, limit: 3, order: .Desc, countryCodes: ["kr", "us"])
                    
                    self.recursiveAppPartnerFriendsCompletion = { [weak self] (friends, error) -> Void in
                        guard let strongSelf = self else { return }
                                                
                        if let friends = friends  {
                            guard let next = PartnerFriendsContext(friends.afterUrl) else { return }
//                            guard let prev = PartnerFriendsContext(friends.beforeUrl) else { return }
                            nextPartnerFriendContext = next
                        }
                        
                        TalkApi.shared.friendsForPartner(context: nextPartnerFriendContext, completion: { (friends, error) in
                            if let error = error {
                                strongSelf.errorHandler(error: error)
                            }
                            else {
                                
                                strongSelf.success()                                
                                strongSelf.recursiveAppPartnerFriendsCompletion(friends, nil)
                            }
                        })
                    }

                    self.recursiveAppPartnerFriendsCompletion(nil, nil)
                }),
                
                Row(title: "friendsForPartner(context:) - PartnerFriendsContext", action: {
                    let nextPartnerFriendContext : PartnerFriendsContext? = PartnerFriendsContext(friendFilter:.None, friendOrder:.Favorite, offset:3, limit:3, order:.Desc, countryCodes: ["kr", "us"])
                    
                    //first
                    TalkApi.shared.friendsForPartner(context:nextPartnerFriendContext) { (friends, error) in
                        if let error = error {
                            self.errorHandler(error: error)
                        }
                        else {
                            self.success()
                            _ = friends
                        }
                    }
                }),
                Row(title: "sendDefaultMessageForPartner(templatable:receiverIdType:receiverIds:) - .Uuid", action: {
                    let title = "카카오톡 링크 4.0(메세지)"
                    let description = "친구에게보내기 티폴트 템플릿 FEED"

                    let feedTemplateJsonStringData =
                        """
                            {
                            "object_type": "feed",
                            "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                            }
                            },
                            "social": {
                            "like_count": 100,
                            "comment_count": 200
                            },
                            "button_title": "바로 확인"
                            }
                            """.data(using: .utf8)!
                
                            if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                                TalkApi.shared.friendsForPartner { [weak self] (friends, error) in
                                    if let error = error {
                                        self?.errorHandler(error: error)
                                    }
                                    else {
                                        self?.success(friends, updateIndicator: false)
                                        guard let friends = friends else {
                                            self?.errorHandler(error: error)
                                            return
                                        }
                                        
                                        if let elements = self?.extractPickerElements(friends: friends) {
                                            let vc = SDKSamplePickerViewController(title: "Friends",
                                                                                     elements: elements,
                                                                                     elementsPickerCompletionHandler: { (selectedIds) in
                                                                                        let friendIds = selectedIds as! [String]
                                                                                        
                                                                                        TalkApi.shared.sendDefaultMessageForPartner(templatable: templatable,
                                                                                                                                    receiverIdType: .Uuid,
                                                                                                                                    receiverIds: friendIds) { (messageSendResult, error) in
                                                                                            if let error = error {
                                                                                                self?.errorHandler(error: error)
                                                                                            }
                                                                                            else {
                                                                                                self?.success()
                                                                                                //do something
                                                                                                let _ = messageSendResult
                                                                                            }
                                                                                        }

                                            })
                                            self?.navigationController?.pushViewController(vc, animated: true)
                                        }
                                    }
                                }
                            }
                }),

                Row(title: "sendDefaultMessageForPartner(templatable:receiverIdType:receiverIds:) - .ChatId", action: {
                    let title = "카카오톡 링크 4.0(메세지)"
                    let description = "친구에게보내기 티폴트 템플릿 FEED"

                    let feedTemplateJsonStringData =
                        """
                            {
                            "object_type": "feed",
                            "content": {
                            "title": "\(title)",
                            "description": "\(description)",
                            "image_url": "http://alpha-api1-kage.kakao.com/dn/cerDB5/ZSb2iRugKx/M4nuZxX823tnK1Mk5yVcv0/kakaolink40_original.png",
                            "link": {
                            "web_url": "http://dev.kakao.com",
                            "mobile_web_url": "http://dev.kakao.com"
                            }
                            },
                            "social": {
                            "like_count": 100,
                            "comment_count": 200
                            },
                            "button_title": "바로 확인"
                            }
                            """.data(using: .utf8)!
//
                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        TalkApi.shared.chatList(completion: { [weak self] (chats, error) in
                            if let error = error {
                                self?.errorHandler(error: error)
                            }
                            else {
                                self?.success(chats, updateIndicator: false)
                                guard let chats = chats else {
                                    self?.errorHandler(error: error)
                                    return
                                }
                                                                
                                if let elements = self?.extractPickerElements(chats: chats) {
                                    let vc = SDKSamplePickerViewController(title: "Chats",
                                                                         elements: elements,
                                                                         elementsPickerCompletionHandler: { (selectedIds) in
                                                                            let chatIds = selectedIds as! [Int64]
                                                                            
                                                                            TalkApi.shared.sendDefaultMessageForPartner(templatable: templatable,
                                                                                                                        receiverIdType: .ChatId,
                                                                                                                        receiverIds: chatIds) { (messageSendResult, error) in
                                                                                if let error = error {
                                                                                    self?.errorHandler(error: error)
                                                                                }
                                                                                else {
                                                                                    self?.success()
                                                                                    //do something
                                                                                    let _ = messageSendResult
                                                                                }
                                                                            }
                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                                }
                            }
                        })
                    }
                }),
                Row(title: "sendCustomMessageForPartner(templateId:receiverIdType:receiverIds:) - .Uuid", action: {
                    TalkApi.shared.friendsForPartner { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            
                            guard let friends = friends else {
                                self?.errorHandler(error: error)
                                return
                            }
                            
                            //uuid
                            if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                        let friendIds = selectedIds as! [String]
                                        
                                        TalkApi.shared.sendCustomMessageForPartner(templateId:AppDelegate.templateId().message,
                                                                                   receiverIdType: .Uuid,
                                                                                   receiverIds: friendIds) { (messageSendResult, error) in
                                            if let error = error {
                                                self?.errorHandler(error: error)
                                            }
                                            else {
                                                self?.success()
                                                //do something
                                                let _ = messageSendResult
                                            }
                                        }
                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    }
                }),
                Row(title: "sendCustomMessageForPartner(templateId:receiverIdType:receiverIds:) - .ChatId", action: {
                    TalkApi.shared.chatList(completion: { [weak self] (chats, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            
                            self?.success(chats, updateIndicator: false)
                            
                            guard let chats = chats else {
                                self?.errorHandler(error: error)
                                return
                            }
                                                            
                            if let elements = self?.extractPickerElements(chats: chats) {
                                let vc = SDKSamplePickerViewController(title: "Chats",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                            let chatIds = selectedIds as! [Int64]
                                            
                                            TalkApi.shared.sendCustomMessageForPartner(templateId:AppDelegate.templateId().message,
                                                                                       receiverIdType: .ChatId,
                                                                                       receiverIds: chatIds) { (messageSendResult, error) in
                                                if let error = error {
                                                    self?.errorHandler(error: error)
                                                }
                                                else {
                                                    self?.success()
                                                    //do something
                                                    let _ = messageSendResult
                                                }
                                            }

                                })
                                self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    })
                }),
                Row(title: "sendCustomMessage(templateId:templateArgs:receiverIdType:receiverIds:) - .Uuid", action: {
                    TalkApi.shared.friendsForPartner { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            
                            guard let friends = friends else {
                                self?.errorHandler(error: error)
                                return
                            }
                            
                            //uuid
                            if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                        let friendIds = selectedIds as! [String]
                                        
                                        TalkApi.shared.sendCustomMessageForPartner(templateId:AppDelegate.templateId().message,
                                                                                   templateArgs:["title":"제목입니다.", "description":"설명입니다."],
                                                                                   receiverIdType: .Uuid,
                                                                                   receiverIds: friendIds) { (messageSendResult, error) in
                                            if let error = error {
                                                self?.errorHandler(error: error)
                                            }
                                            else {
                                                self?.success()
                                                //do something
                                                let _ = messageSendResult
                                            }
                                        }
                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    }
                }),
                Row(title: "sendCustomMessageForPartner(templateId:templateArgs:receiverIdType:receiverIds:) - .ChatId", action: {
                    TalkApi.shared.chatList(completion: { [weak self] (chats, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            
                            self?.success(chats, updateIndicator: false)
                            
                            guard let chats = chats else {
                                self?.errorHandler(error: error)
                                return
                            }
                                                            
                            if let elements = self?.extractPickerElements(chats: chats) {
                                let vc = SDKSamplePickerViewController(title: "Chats",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                            let chatIds = selectedIds as! [Int64]
                                            
                                            TalkApi.shared.sendCustomMessageForPartner(templateId:AppDelegate.templateId().message,
                                                                                       receiverIdType: .ChatId,
                                                                                       receiverIds: chatIds) { (messageSendResult, error) in
                                                if let error = error {
                                                    self?.errorHandler(error: error)
                                                }
                                                else {
                                                    self?.success()
                                                    //do something
                                                    let _ = messageSendResult
                                                }
                                            }

                                })
                                self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    })
                }),
                Row(title: "sendScrapMessageForPartner(requestUrl:receiverUuids:) - .Uuid", action: {
                    TalkApi.shared.friendsForPartner { [weak self] (friends, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success(friends, updateIndicator: false)
                            
                            guard let friends = friends else {
                                self?.errorHandler(error: error)
                                return
                            }
                            
                            //uuid
                            if let elements = self?.extractPickerElements(friends: friends) {
                                    let vc = SDKSamplePickerViewController(title: "Friends",
                                                                             elements: elements,
                                                                             elementsPickerCompletionHandler: { (selectedIds) in
                                        let friendIds = selectedIds as! [String]
                                                                                
                                        TalkApi.shared.sendScrapMessageForPartner(requestUrl:"https://developers.kakao.com",
                                                                                  receiverIdType: .Uuid,
                                                                                  receiverIds: friendIds) { (messageSendResult, error) in
                                            if let error = error {
                                                self?.errorHandler(error: error)
                                            }
                                            else {
                                                self?.success()
                                                //do something
                                                let _ = messageSendResult
                                            }
                                        }
                                    })
                                    self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    }
                }),
                Row(title: "sendScrapMessageForPartner(requestUrl:templateId:receiverChatIds:) - .ChatId", action: {
                    TalkApi.shared.chatList(completion: { [weak self] (chats, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            
                            self?.success(chats, updateIndicator: false)
                            
                            guard let chats = chats else {
                                self?.errorHandler(error: error)
                                return
                            }
                                                            
                            if let elements = self?.extractPickerElements(chats: chats) {
                                let vc = SDKSamplePickerViewController(title: "Chats",
                                                                       elements: elements,
                                                                       elementsPickerCompletionHandler: { (selectedIds) in
                                            let chatIds = selectedIds as! [Int64]
                                            
                                            TalkApi.shared.sendScrapMessageForPartner(requestUrl: "https://developers.kakao.com",
                                                                                      templateId:AppDelegate.templateId().message,
                                                                                      receiverIdType: .ChatId,
                                                                                      receiverIds: chatIds) { (messageSendResult, error) in
                                                if let error = error {
                                                    self?.errorHandler(error: error)
                                                }
                                                else {
                                                    self?.success()
                                                    //do something
                                                    let _ = messageSendResult
                                                }
                                            }

                                })
                                self?.navigationController?.pushViewController(vc, animated: true)
                            }
                        }
                    })
                }),
                
            ]),
            // section talk end =============================================================================================
            // MARK: - section link start =============================================================================================
            SectionOfRow(header: "SHARE - Web Sharer", items: [
                Row(title: "shareDefault(targetAppKey:templatable:) - webSharer", action: {
                    let feedTemplateJsonStringData =
                      """
                      {
                          "object_type": "feed",
                          "content": {
                              "title": "딸기 치즈 케익",
                              "description": "#케익 #딸기 #삼평동 #카페 #분위기 #소개팅",
                              "image_url": "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                              "link": {
                                  "mobile_web_url": "https://developers.kakao.com",
                                  "web_url": "https://developers.kakao.com"
                              }
                          },
                          "item_content" : {
                                "profile_text" :"Kakao",
                                "profile_image_url" :"http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_url" : "http://mud-kage.kakao.co.kr/dn/Q2iNx/btqgeRgV54P/VLdBs9cvyn8BJXB3o7N8UK/kakaolink40_original.png",
                                "title_image_text" :"Cheese cake",
                                "title_image_category" : "cake",
                                "items" : [
                                                {
                                                    "item" :"cake1",
                                                    "item_op" : "1000원",
                                                },
                                                {
                                                    "item" :"cake2",
                                                    "item_op" : "2000원",
                                                },
                                                {
                                                    "item" :"cake3",
                                                    "item_op" : "3000원",
                                                },
                                                {
                                                    "item" :"cake4",
                                                    "item_op" : "4000원",
                                                },
                                                {
                                                    "item" :"cake5",
                                                    "item_op" : "5000원",
                                                }
                                            ],
                                            "sum" :"total",
                                            "sum_op" : "15000원"
                          },
                          "social": {
                              "comment_count": 45,
                              "like_count": 286,
                              "shared_count": 845
                          },
                          "buttons": [
                              {
                                  "title": "웹으로 보기",
                                  "link": {
                                      "mobile_web_url": "https://developers.kakao.com",
                                      "web_url": "https://developers.kakao.com"
                                  }
                              },
                              {
                                  "title": "앱으로 보기",
                                  "link": {
                                      "android_execution_params": "key1=value1&key2=value2",
                                      "ios_execution_params": "key1=value1&key2=value2"
                                  }
                              }
                          ]
                      }
                      """.data(using: .utf8)!
                    //출처용 앱키설정 필요
                    let targetAppKey = "cd9b86b063c3037851a4671c1712c01b"

                    if let templatable = try? SdkJSONDecoder.custom.decode(FeedTemplate.self, from: feedTemplateJsonStringData) {
                        if let url = ShareApi.shared.makeDefaultUrl(targetAppKey:targetAppKey, templatable: templatable) {
                            print("=============================sharer default url: \(url)")
                            self.safariViewController = SFSafariViewController(url: url)
                            self.safariViewController?.modalTransitionStyle = .crossDissolve
                            self.safariViewController?.modalPresentationStyle = .overCurrentContext
                            self.present(self.safariViewController!, animated: true) {
                                print("웹 present success")
                            }
                        }
                    }
                }),
                Row(title: "shareScrap(targetAppKey:requestUrl:) - webSharer", action: {
                    //출처용 앱키설정 필요
                    let targetAppKey = "cd9b86b063c3037851a4671c1712c01b"
                    
                    if let url = ShareApi.shared.makeScrapUrl(targetAppKey:targetAppKey, requestUrl: "https://developers.kakao.com") {
                        print("=============================sharer scrap url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                }),
                Row(title: "shareCustom(targetAppKey:templateId:,templateArgs:) - webSharer", action: {
                    //출처용 앱키설정 필요
                    let targetAppKey = "cd9b86b063c3037851a4671c1712c01b"
                    
                    if let url = ShareApi.shared.makeCustomUrl(targetAppKey:targetAppKey, templateId:AppDelegate.templateId().link, templateArgs:["title":"제목입니다.", "description":"설명입니다."]) {
                        print("=============================sharer custom url: \(url)")
                        self.safariViewController = SFSafariViewController(url: url)
                        self.safariViewController?.modalTransitionStyle = .crossDissolve
                        self.safariViewController?.modalPresentationStyle = .overCurrentContext
                        self.present(self.safariViewController!, animated: true) {
                            print("웹 present success")
                        }
                    }
                })
            ]),
            // section link start =============================================================================================
            
            // MARK: - section etc start =============================================================================================
            SectionOfRow(header: "ETC", items: [
                Row(title: "token() usage", action: {
                    AuthApi.shared.token(code: "#발급받은코드#" ) { (token, error) in
                        if let error = error {
                            print(error)
                        }
                        else {
                            //success
                            //토큰은 SDK 내부 저장소에 저장됩니다.(이후 SDK사용은 기존과 똑같이 하면됩니다.)
                        }
                    }
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[view:login],[email:a@kakao.com]", action: {
                    UserApi.shared.loginWithKakaoAccount(accountParameters: ["view":"login", "email":"a@kakao.com"]) { [weak self] (token, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                            
                            print("token : \(String(describing: token))")
                        }
                    }
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[view:signup]", action: {
                    UserApi.shared.loginWithKakaoAccount(accountParameters: ["view":"signup"]) { [weak self] (token, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                            
                            print("token : \(String(describing: token))")
                        }
                    }
                }),
                Row(title: "loginWithKakaoAccountAtAccountPage(accountParameters:)-[:]", action: {
                    UserApi.shared.loginWithKakaoAccount(accountParameters: [:]) { [weak self] (token, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            //do something
                            
                            print("token : \(String(describing: token))")
                        }
                    }
                }),
                Row(title: "loginWithKakaoTalk(service_terms: abcd)", action: {
                    UserApi.shared.loginWithKakaoTalk(serviceTerms: ["abcd"]) { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            let _ = oauthToken
                            //do something
                        }
                    }
                }),
                Row(title: "loginWithKakaoTalk(service_terms: 가나다라)", action: {
                    UserApi.shared.loginWithKakaoTalk(serviceTerms: ["가나다라"]) { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            let _ = oauthToken
                            //do something
                        }
                    }
                }),
                Row(title: "loginWithKakaoTalk(service_terms: ABC?가=나&다)", action: {
                    UserApi.shared.loginWithKakaoTalk(serviceTerms: ["ABC?가=나&다"]) { [weak self] (oauthToken, error) in
                        if let error = error {
                            self?.errorHandler(error: error)
                        }
                        else {
                            self?.success()
                            
                            let _ = oauthToken
                            //do something
                        }
                    }
                }),
                Row(title: "loginWithKakaoTalk() - URL Decoding Error (only real)", action: {
                    
                    // 중복 디코딩 - https://jira.daumkakao.com/browse/OTF-2959
                    
                    let url = URL(string: "kakaokompassauth://authorize?client_id=8c55ce417911a5515368ee5e87021928&redirect_uri=http://192.168.10.104:8000/test/target/manual.html&params=%7B%22client_id%22%3A%228c55ce417911a5515368ee5e87021928%22%2C%22state%22%3A%22%7B%5C%22redirectUri%5C%22%3A%20%5C%22http%253A%252F%252F10.123.4.56%253A1234%252Fkakao-login.do%253Fi_sType%253DM%2526i_sReturnUrl%253D%252Fmweb%252Fshop%252Fproduct%252FproductView.do%2526i_sReturnParam%253D%2526i_sFlagLoginSucc%253DY%2526s_fsFlag%253DY%2526i_sProductcd%253D13213123%2526i_sFlagMileShop%253D%5C%22%2C%20%5C%22siteCd%5C%22%3A%5C%22030%5C%22%2C%20%5C%22clientId%5C%22%3A%20%5C%22722ac15b90f0441a86d55eb051bf8711%5C%22%2C%5C%22channel%5C%22%3A%5C%22WEB%5C%22%2C%5C%22snsType%5C%22%3A%5C%22KA%5C%22%2C%5C%22kakaoChannel%5C%22%3A%5C%22_clpYT%5C%22%7D%22%2C%22redirect_uri%22%3A%22http%3A%2F%2F192.168.10.104%3A8000%2Ftest%2Ftarget%2Fmanual.html%22%2C%22response_type%22%3A%22code%22%2C%22auth_tran_id%22%3A%22ub7byckqqkp8c55ce417911a5515368ee5e87021928k8tjmgfw%22%2C%22is_popup%22%3Atrue%7D&headers=%7B%22ka%22%3A%22sdk%2F1.38.0-DEV%20os%2Fjavascript%20lang%2Fko%20device%2FMacIntel%20origin%2Fhttp%253A%252F%252F192.168.10.104%253A8000%22%7D&web=https://kauth.kakao.com/oauth/authorize?client_id=8c55ce417911a5515368ee5e87021928&state=%7B%22redirectUri%22%3A%20%22http%253A%252F%252F10.123.4.56%253A1234%252Fkakao-login.do%253Fi_sType%253DM%2526i_sReturnUrl%253D%252Fmweb%252Fshop%252Fproduct%252FproductView.do%2526i_sReturnParam%253D%2526i_sFlagLoginSucc%253DY%2526s_fsFlag%253DY%2526i_sProductcd%253D13213123%2526i_sFlagMileShop%253D%22%2C%20%22siteCd%22%3A%22030%22%2C%20%22clientId%22%3A%20%22722ac15b90f0441a86d55eb051bf8711%22%2C%22channel%22%3A%22WEB%22%2C%22snsType%22%3A%22KA%22%2C%22kakaoChannel%22%3A%22_clpYT%22%7D&redirect_uri=http%3A%2F%2F192.168.10.104%3A8000%2Ftest%2Ftarget%2Fmanual.html&response_type=code&auth_tran_id=ub7byckqqkp8c55ce417911a5515368ee5e87021928k8tjmgfw")!
                    print(url)
                    
                    UIApplication.shared.open(url, options: [:], completionHandler: nil)
                }),
            ])
            // section etc start =============================================================================================
        ]
    }
}
